# -*- coding: utf-8 -*-
"""
comm.py
"""

import numpy as np
from numba import jit
from mpi4py import MPI

# 入力データを全プロセスで共有する
def broadcast(Parm,
    Nx, Ny, Nz, Xn, Yn, Zn,
    iMaterial, fMaterial, iGeometry, fGeometry,
    iFeed, fFeed, iPoint, fPoint, iInductor, fInductor, Freq1, Freq2):

    i_buf = None
    d_buf = None
    i_num = np.zeros(1, 'i4')
    d_num = np.zeros(1, 'i4')

    # rootに変数を格納する
    if Parm['comm_rank'] == 0:
        #print(iMaterial.size, iGeometry.size)
        #print(fMaterial.size, fGeometry.size)
        #print(Xn.size, Yn.size, Zn.size)
        #print(iFeed.size, fFeed.size)
        #print(Freq1.size, Freq2.size)

        # データ数
        i_num[0] = 20 \
            + iMaterial.size + iGeometry.size \
            + iFeed.size + iPoint.size + iInductor.size
        d_num[0] = 8 \
            + Xn.size + Yn.size + Zn.size + fMaterial.size + fGeometry.size \
            + fFeed.size + fPoint.size + fInductor.size \
            + Freq1.size + Freq2.size

        # alloc
        i_buf = np.zeros(i_num, 'i4')
        d_buf = np.zeros(d_num, 'f8')

        i_id = 0
        d_id = 0

        # 整数データ(20個)
        i_buf[i_id] = Nx;                   i_id += 1
        i_buf[i_id] = Ny;                   i_id += 1
        i_buf[i_id] = Nz;                   i_id += 1
        i_buf[i_id] = fMaterial.shape[0];   i_id += 1
        i_buf[i_id] = fGeometry.shape[0];   i_id += 1
        i_buf[i_id] = fFeed.shape[0];       i_id += 1
        i_buf[i_id] = fPoint.shape[0];      i_id += 1
        i_buf[i_id] = fInductor.shape[0];   i_id += 1
        i_buf[i_id] = len(Freq1);           i_id += 1
        i_buf[i_id] = len(Freq2);           i_id += 1
        i_buf[i_id] = Parm['source'];       i_id += 1
        i_buf[i_id] = Parm['planewave'][2]; i_id += 1
        i_buf[i_id] = Parm['abc'][0];       i_id += 1
        i_buf[i_id] = Parm['abc'][1];       i_id += 1
        i_buf[i_id] = Parm['pbc'][0];       i_id += 1
        i_buf[i_id] = Parm['pbc'][1];       i_id += 1
        i_buf[i_id] = Parm['pbc'][2];       i_id += 1
        i_buf[i_id] = Parm['prop'];         i_id += 1
        i_buf[i_id] = Parm['solver'][0];    i_id += 1
        i_buf[i_id] = Parm['solver'][1];    i_id += 1

        # 実数データ(8個)
        d_buf[d_id] = Parm['planewave'][0]; d_id += 1
        d_buf[d_id] = Parm['planewave'][1]; d_id += 1
        d_buf[d_id] = Parm['rfeed'];        d_id += 1
        d_buf[d_id] = Parm['abc'][2];       d_id += 1
        d_buf[d_id] = Parm['abc'][3];       d_id += 1
        d_buf[d_id] = Parm['solver'][2];    d_id += 1
        d_buf[d_id] = Parm['dt'];           d_id += 1
        d_buf[d_id] = Parm['tw'];           d_id += 1

        for i in range(len(Xn)):
            d_buf[d_id] = Xn[i]; d_id += 1
        for j in range(len(Yn)):
            d_buf[d_id] = Yn[j]; d_id += 1
        for k in range(len(Zn)):
            d_buf[d_id] = Zn[k]; d_id += 1

        for n in range(iMaterial.shape[0]):
            i_buf[i_id] = iMaterial[n]; i_id += 1
        for n in range(fMaterial.shape[0]):
            for m in range(fMaterial.shape[1]):
                d_buf[d_id] = fMaterial[n, m]; d_id += 1

        for n in range(iGeometry.shape[0]):
            for m in range(iGeometry.shape[1]):
                i_buf[i_id] = iGeometry[n, m]; i_id += 1
        for n in range(fGeometry.shape[0]):
            for m in range(fGeometry.shape[1]):
                d_buf[d_id] = fGeometry[n, m]; d_id += 1

        for n in range(iFeed.shape[0]):
            for m in range(iFeed.shape[1]):
                i_buf[i_id] = iFeed[n, m]; i_id += 1
        for n in range(fFeed.shape[0]):
            for m in range(fFeed.shape[1]):
                d_buf[d_id] = fFeed[n, m]; d_id += 1

        for n in range(iPoint.shape[0]):
            for m in range(iPoint.shape[1]):
                i_buf[i_id] = iPoint[n, m]; i_id += 1
        for n in range(fPoint.shape[0]):
            for m in range(fPoint.shape[1]):
                d_buf[d_id] = fPoint[n, m]; d_id += 1

        for n in range(iInductor.shape[0]):
            for m in range(iInductor.shape[1]):
                i_buf[i_id] = iInductor[n, m]; i_id += 1
        for n in range(fInductor.shape[0]):
            for m in range(fInductor.shape[1]):
                d_buf[d_id] = fInductor[n, m]; d_id += 1

        for _, freq1 in enumerate(Freq1):
            d_buf[d_id] = freq1; d_id += 1

        for _, freq2 in enumerate(Freq2):
            d_buf[d_id] = freq2; d_id += 1

        # check
        assert(i_id == i_num[0])
        assert(d_id == d_num[0])

    # broadcast (root to non-root)
    MPI.COMM_WORLD.Bcast(i_num)
    MPI.COMM_WORLD.Bcast(d_num)
    #print(Parm['comm_size'], Parm['comm_rank'], i_num[0], d_num[0])

    # alloc
    if Parm['comm_rank'] > 0:
        i_buf = np.zeros(i_num[0], 'i4')
        d_buf = np.zeros(d_num[0], 'f8')

    MPI.COMM_WORLD.Bcast(i_buf)
    MPI.COMM_WORLD.Bcast(d_buf)

    # 受信したデータを変数に代入する (非root)
    if Parm['comm_rank'] > 0:
        i_id = 0
        d_id = 0

        # パラメーター初期化
        Parm['source']     = 0
        Parm['planewave']  = [0] * 3
        Parm['rfeed']      = 0
        Parm['abc']        = [0] * 4
        Parm['pbc']        = [0] * 3
        Parm['prop']       = 0
        Parm['solver']     = [1000, 50, 1e-3]
        Parm['dt']         = 0
        Parm['tw']         = 0

        # 整数データ(20個)
        Nx = i_buf[i_id];                             i_id += 1
        Ny = i_buf[i_id];                             i_id += 1
        Nz = i_buf[i_id];                             i_id += 1
        iMaterial = np.zeros( i_buf[i_id],     'i4'); i_id += 1
        iGeometry = np.zeros((i_buf[i_id], 2), 'i4'); i_id += 1
        iFeed     = np.zeros((i_buf[i_id], 4), 'i4'); i_id += 1
        iPoint    = np.zeros((i_buf[i_id], 4), 'i4'); i_id += 1
        iInductor = np.zeros((i_buf[i_id], 4), 'i4'); i_id += 1
        Freq1     = np.zeros( i_buf[i_id],     'f8'); i_id += 1
        Freq2     = np.zeros( i_buf[i_id],     'f8'); i_id += 1
        Parm['source']       = i_buf[i_id];           i_id += 1
        Parm['planewave'][2] = i_buf[i_id];           i_id += 1
        Parm['abc'][0]       = i_buf[i_id];           i_id += 1
        Parm['abc'][1]       = i_buf[i_id];           i_id += 1
        Parm['pbc'][0]       = i_buf[i_id];           i_id += 1
        Parm['pbc'][1]       = i_buf[i_id];           i_id += 1
        Parm['pbc'][2]       = i_buf[i_id];           i_id += 1
        Parm['prop']         = i_buf[i_id];           i_id += 1
        Parm['solver'][0]    = i_buf[i_id];           i_id += 1
        Parm['solver'][1]    = i_buf[i_id];           i_id += 1

        fMaterial = np.zeros((iMaterial.shape[0], 11), 'f8')
        fGeometry = np.zeros((iGeometry.shape[0],  8), 'f8')
        fFeed     = np.zeros((iFeed.shape[0],      9), 'f8')
        fPoint    = np.zeros((iPoint.shape[0],     6), 'f8')
        fInductor = np.zeros((iInductor.shape[0], 10), 'f8')

        # 実数データ(8個)
        Parm['planewave'][0] = d_buf[d_id]; d_id += 1
        Parm['planewave'][1] = d_buf[d_id]; d_id += 1
        Parm['rfeed']        = d_buf[d_id]; d_id += 1
        Parm['abc'][2]       = d_buf[d_id]; d_id += 1
        Parm['abc'][3]       = d_buf[d_id]; d_id += 1
        Parm['solver'][2]    = d_buf[d_id]; d_id += 1
        Parm['dt']           = d_buf[d_id]; d_id += 1
        Parm['tw']           = d_buf[d_id]; d_id += 1
        #print(Parm['comm_rank'], Parm)

        Xn = np.zeros(Nx + 1, 'f8')
        Yn = np.zeros(Ny + 1, 'f8')
        Zn = np.zeros(Nz + 1, 'f8')
        for i in range(Nx + 1):
            Xn[i] = d_buf[d_id]; d_id += 1
        for j in range(Ny + 1):
            Yn[j] = d_buf[d_id]; d_id += 1
        for k in range(Nz + 1):
            Zn[k] = d_buf[d_id]; d_id += 1
        #print(Parm['comm_rank'], Xn, Yn, Zn)

        for n in range(iMaterial.shape[0]):
            iMaterial[n] = i_buf[i_id]; i_id += 1
        for n in range(fMaterial.shape[0]):
            for m in range(fMaterial.shape[1]):
                fMaterial[n, m] = d_buf[d_id]; d_id += 1
        #print(Parm['comm_rank'], iMaterial, fMaterial)

        for n in range(iGeometry.shape[0]):
            for m in range(iGeometry.shape[1]):
                iGeometry[n, m] = i_buf[i_id]; i_id += 1
        for n in range(fGeometry.shape[0]):
            for m in range(fGeometry.shape[1]):
                fGeometry[n, m] = d_buf[d_id]; d_id += 1
        #print(Parm['comm_rank'], iGeometry, fGeometry)

        for n in range(iFeed.shape[0]):
            for m in range(iFeed.shape[1]):
                iFeed[n, m] = i_buf[i_id]; i_id += 1
        for n in range(fFeed.shape[0]):
            for m in range(fFeed.shape[1]):
                fFeed[n, m] = d_buf[d_id]; d_id += 1

        for n in range(iPoint.shape[0]):
            for m in range(iPoint.shape[1]):
                iPoint[n, m] = i_buf[i_id]; i_id += 1
        for n in range(fPoint.shape[0]):
            for m in range(fPoint.shape[1]):
                fPoint[n, m] = d_buf[d_id]; d_id += 1

        for n in range(iInductor.shape[0]):
            for m in range(iInductor.shape[1]):
                iInductor[n, m] = i_buf[i_id]; i_id += 1
        for n in range(fInductor.shape[0]):
            for m in range(fInductor.shape[1]):
                fInductor[n, m] = d_buf[d_id]; d_id += 1

        for n in range(len(Freq1)):
            Freq1[n] = d_buf[d_id]; d_id += 1

        for n in range(len(Freq2)):
            Freq2[n] = d_buf[d_id]; d_id += 1
        #print(Parm['comm_rank'], Freq1, Freq2)

        # check
        assert(i_id == i_num[0])
        assert(d_id == d_num[0])

    # free
    i_buf = None
    d_buf = None

    return \
    Nx, Ny, Nz, Xn, Yn, Zn, \
    iMaterial, fMaterial, iGeometry, fGeometry, \
    iFeed, fFeed, iPoint, fPoint, iInductor, fInductor, Freq1, Freq2

"""
# X境界のHy/Hzを共有する (Send + Recv)
def share_boundary(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    buf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # +X境界送信
    if comm_rank < comm_size - 1:
        buf[: ls] = Hy[offset_hy[2]: offset_hy[2] + length_hy]
        buf[ls: ] = Hz[offset_hz[2]: offset_hz[2] + length_hz]
        MPI.COMM_WORLD.Send(buf, comm_rank + 1)

    # -X境界受信
    if comm_rank > 0:
        MPI.COMM_WORLD.Recv(buf, source = comm_rank - 1)
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = buf[: ls]
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = buf[ls: ]

    # -X境界送信
    if comm_rank > 0:
        buf[: ls] = Hy[offset_hy[1]: offset_hy[1] + length_hy]
        buf[ls: ] = Hz[offset_hz[1]: offset_hz[1] + length_hz]
        MPI.COMM_WORLD.Send(buf, comm_rank - 1)

    # +X境界受信
    if comm_rank < comm_size - 1:
        MPI.COMM_WORLD.Recv(buf, source = comm_rank + 1)
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = buf[: ls]
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = buf[ls: ]
"""

# X境界のHy/Hzを共有する (Sendrecv)
# こちらを採用する
def share_boundary(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    sendbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    recvbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # -X境界
    if comm_rank > 0:
        # copy to buffer
        sendbuf[: ls] = Hy[offset_hy[1]: offset_hy[1] + length_hy]
        sendbuf[ls: ] = Hz[offset_hz[1]: offset_hz[1] + length_hz]

        # MPI
        dest = comm_rank - 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = recvbuf[: ls]
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = recvbuf[ls: ]

    # +X境界
    if comm_rank < comm_size - 1:
        # copy to buffer
        sendbuf[: ls] = Hy[offset_hy[2]: offset_hy[2] + length_hy]
        sendbuf[ls: ] = Hz[offset_hz[2]: offset_hz[2] + length_hz]

        # MPI
        dest = Parm['comm_rank'] + 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = recvbuf[: ls]
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = recvbuf[ls: ]

"""
# PBC : -X/+X境界のHy/Hzを共有する (Send + Recv)
def pbcx(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    buf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # -X境界送信
    if comm_rank == 0:
        buf[: ls] = Hy[offset_hy[1]: offset_hy[1] + length_hy]
        buf[ls: ] = Hz[offset_hz[1]: offset_hz[1] + length_hz]
        MPI.COMM_WORLD.Send(buf, comm_size - 1)

    # +X境界受信
    if comm_rank == comm_size - 1:
        MPI.COMM_WORLD.Recv(buf, source = 0)
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = buf[: ls]
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = buf[ls: ]

    # +X境界送信
    if comm_rank == comm_size - 1:
        buf[: ls] = Hy[offset_hy[2]: offset_hy[2] + length_hy]
        buf[ls: ] = Hz[offset_hz[2]: offset_hz[2] + length_hz]
        MPI.COMM_WORLD.Send(buf, 0)

    # -X境界受信
    if comm_rank == 0:
        MPI.COMM_WORLD.Recv(buf, source = comm_size - 1)
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = buf[: ls]
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = buf[ls: ]
"""

# PBC : -X/+X境界のHy/Hzを共有する (Sendrecv)
# こちらを採用する
def pbcx(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    sendbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    recvbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # -X boundary
    if comm_rank == 0:
        # copy to buffer
        sendbuf[: ls] = Hy[offset_hy[1]: offset_hy[1] + length_hy]
        sendbuf[ls: ] = Hz[offset_hz[1]: offset_hz[1] + length_hz]

        # MPI
        dest = comm_size - 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = recvbuf[: ls]
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = recvbuf[ls: ]

    # +X boundary
    elif comm_rank == comm_size - 1:
        # copy to buffer
        sendbuf[: ls] = Hy[offset_hy[2]: offset_hy[2] + length_hy]
        sendbuf[ls: ] = Hz[offset_hz[2]: offset_hz[2] + length_hz]

        # MPI
        dest = 0
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = recvbuf[: ls]
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = recvbuf[ls: ]


# 和(スカラー, 全プロセスで共有)
def sum_scalar(var):
    sendbuf = np.array([var], 'f8')
    recvbuf = np.zeros(1, 'f8')
    MPI.COMM_WORLD.Allreduce(sendbuf, recvbuf)

    return recvbuf[0]

# 全領域の電磁界をrootに集める
def gather_near3d(Parm, Freq2,
    cEx, cEy, cEz, cHx, cHy, cHz,
    iMin, iMax, jMin, jMax, kMin, kMax, Ni, Nj, Nk, N0, NN, NN_root):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']
    #print(comm_size, comm_rank)

    if (NN <= 0) or (len(Freq2) <= 0):
        return

    g_cEx = g_cEy = g_cEz = g_cHx = g_cHy = g_cHz = None
    
    # root : 全領域配列
    
    if comm_rank == 0:
        fsize = NN * len(Freq2)
        g_cEx = np.zeros(fsize, 'c8')
        g_cEy = np.zeros(fsize, 'c8')
        g_cEz = np.zeros(fsize, 'c8')
        g_cHx = np.zeros(fsize, 'c8')
        g_cHy = np.zeros(fsize, 'c8')
        g_cHz = np.zeros(fsize, 'c8')

    
    for ifreq in range(len(Freq2)):

        if comm_rank == 0:
            # root

            # === self copy (0->0) ===

            g_n = ifreq * NN
            n   = ifreq * NN_root
            g_cEx[g_n: g_n + NN_root] = cEx[n: n + NN_root].copy()
            g_cEy[g_n: g_n + NN_root] = cEy[n: n + NN_root].copy()
            g_cEz[g_n: g_n + NN_root] = cEz[n: n + NN_root].copy()
            g_cHx[g_n: g_n + NN_root] = cHx[n: n + NN_root].copy()
            g_cHy[g_n: g_n + NN_root] = cHy[n: n + NN_root].copy()
            g_cHz[g_n: g_n + NN_root] = cHz[n: n + NN_root].copy()

            # === 受信 ===

            for rank in range(1, comm_size):

                # 受信
                irecv = np.zeros(11, 'i4')
                MPI.COMM_WORLD.Recv(irecv, source=rank)

                imin, imax, jmin, jmax, kmin, kmax = irecv[0:6]
                ni, nj, nk, n0, nn = irecv[6:11]  # E/H 6成分共通
                #print(imin, imax, jmin, jmax, kmin, kmax, ni, nj, nk, n0, nn)
                
                # 受信用バッファー
                recv = np.zeros(nn, 'c8')

                # copy to global array

                # Ex
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cEx, recv,
                    imin + 0, imax + 0, jmin + 0, jmax + 1, kmin + 0, kmax + 1,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """f
                for i in range(imin, imax + 0):
                    for j in range(jmin, jmax + 1):
                        for k in range(kmin, kmax + 1):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cEx[g_n] = recv[n]"""

                # Ey
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cEy, recv,
                    imin + 0, imax + 1, jmin + 0, jmax + 0, kmin + 0, kmax + 1,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """
                for i in range(imin, imax + 1):
                    for j in range(jmin, jmax + 0):
                        for k in range(kmin, kmax + 1):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cEy[g_n] = recv[n]"""

                # Ey
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cEz, recv,
                    imin + 0, imax + 1, jmin + 0, jmax + 1, kmin + 0, kmax + 0,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """
                for i in range(imin, imax + 1):
                    for j in range(jmin, jmax + 1):
                        for k in range(kmin, kmax + 0):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cEz[g_n] = recv[n]"""

                # Hx
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cHx, recv,
                    imin - 0, imax + 1, jmin - 1, jmax + 1, kmin - 1, kmax + 1,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """
                for i in range(imin - 0, imax + 1):
                    for j in range(jmin - 1, jmax + 1):
                        for k in range(kmin - 1, kmax + 1):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cHx[g_n] = recv[n]"""

                # Hy
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cHy, recv,
                    imin - 1, imax + 1, jmin - 0, jmax + 1, kmin - 1, kmax + 1,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """
                for i in range(imin - 1, imax + 1):
                    for j in range(jmin - 0, jmax + 1):
                        for k in range(kmin - 1, kmax + 1):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cHy[g_n] = recv[n]"""

                # Hz
                MPI.COMM_WORLD.Recv(recv, source=rank)
                _gather(g_cHz, recv,
                    imin - 1, imax + 1, jmin - 1, jmax + 1, kmin - 0, kmax + 1,
                    Ni, Nj, Nk, N0, ni, nj, nk, n0, ifreq * NN)
                """
                for i in range(imin - 1, imax + 1):
                    for j in range(jmin - 1, jmax + 1):
                        for k in range(kmin - 0, kmax + 1):
                            g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + (ifreq * NN)
                            n   = (ni * i) + (nj * j) + (nk * k) + n0
                            g_cHz[g_n] = recv[n]"""

                # free
                recv = None

        else:
            # 非root : rootに送信する

            # indexを送信する(11個)
            isend = np.array([iMin, iMax, jMin, jMax, kMin, kMax, \
                              Ni, Nj, Nk, N0, NN]).astype('i4')
            MPI.COMM_WORLD.Send(isend, 0)

            # データ本体を送信する
            #print(NN)
            n0 = (ifreq + 0) * NN
            n1 = (ifreq + 1) * NN
            MPI.COMM_WORLD.Send(cEx[n0: n1], 0)
            MPI.COMM_WORLD.Send(cEy[n0: n1], 0)
            MPI.COMM_WORLD.Send(cEz[n0: n1], 0)
            MPI.COMM_WORLD.Send(cHx[n0: n1], 0)
            MPI.COMM_WORLD.Send(cHy[n0: n1], 0)
            MPI.COMM_WORLD.Send(cHz[n0: n1], 0)
    
    
    # root : ポインターをコピーする
    if comm_rank == 0:
        #cEx = cEy = cEz = cHx = cHy = cHz = None
        cEx = g_cEx
        cEy = g_cEy
        cEz = g_cEz
        cHx = g_cHx
        cHy = g_cHy
        cHz = g_cHz

    return cEx, cEy, cEz, cHx, cHy, cHz

# 
@jit(cache=True, nopython=True)
def _gather(g_f, f,
    i_min, i_max, j_min, j_max, k_min, k_max,
    Ni, Nj, Nk, N0, ni, nj, nk, n0, g_n0):

    for i in range(i_min, i_max):
        for j in range(j_min, j_max):
            for k in range(k_min, k_max):
                g_n = (Ni * i) + (Nj * j) + (Nk * k) + N0 + g_n0
                n   = (ni * i) + (nj * j) + (nk * k) + n0
                g_f[g_n] = f[n]

# send feed waveform to root process
def gather_feed(Parm, Nx, iMin, iMax, iFeed, VFeed, IFeed):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    count = Parm['solver'][0] + 1

    for ifeed in range(iFeed.shape[0]):
        i = iFeed[ifeed, 1]
        # 非rootにあるときのみ通信する
        if   (comm_rank == 0) and \
            not _mpi_proc(comm_size, comm_rank, Nx, iMin, iMax, i):
            MPI.COMM_WORLD.Recv(VFeed[ifeed, :count], source=MPI.ANY_SOURCE)
            MPI.COMM_WORLD.Recv(IFeed[ifeed, :count], source=MPI.ANY_SOURCE)
        elif (comm_rank  > 0) and \
            _mpi_proc(comm_size, comm_rank, Nx, iMin, iMax, i):
            MPI.COMM_WORLD.Send(VFeed[ifeed, :count], 0)
            MPI.COMM_WORLD.Send(IFeed[ifeed, :count], 0)

        MPI.COMM_WORLD.Barrier()

# send point waveform to root process
def gather_point(Parm, Nx, iMin, iMax, iPoint, VPoint):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    count = Parm['solver'][0] + 1

    for ipoint in range(iPoint.shape[0]):
        i = iPoint[ipoint, 1]
        # 非rootにあるときのみ通信する
        if   (comm_rank == 0) and \
            not _mpi_proc(comm_size, comm_rank, Nx, iMin, iMax, i):
            MPI.COMM_WORLD.Recv(VPoint[ipoint, :count], source=MPI.ANY_SOURCE)
        elif (comm_rank  > 0) and \
            _mpi_proc(comm_size, comm_rank, Nx, iMin, iMax, i):
            MPI.COMM_WORLD.Send(VPoint[ipoint, :count], 0)

        MPI.COMM_WORLD.Barrier()

# MPI buffer size
def buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    # ABC層数
    #lx = 1 if (Parm['abc'][0] == 0) else Parm['abc'][1]
    ly = 1 if (Parm['abc'][0] == 0) else Parm['abc'][1]
    lz = 1 if (Parm['abc'][0] == 0) else Parm['abc'][1]

    jmin_hy = -ly + 1
    jmin_hz = -ly + 0
    kmin_hy = -lz + 0
    kmin_hz = -lz + 1
    #print(Parm['comm_rank'], jmin_hy, jmin_hz, kmin_hy, kmin_hz)

    jmax_hy = Ny + ly
    jmax_hz = Ny + ly
    kmax_hy = Nz + lz
    kmax_hz = Nz + lz
    #print(Parm['comm_rank'], jmax_hy, jmax_hz, kmax_hy, kmax_hz)

    # [0] : -X boundary - 1 (recv)
    # [1] : -X boundary (send)
    # [2] : +X boundary (send)
    # [3] : +X boundary + 1 (recv)

    offset_hy = np.zeros(4, 'i4')
    offset_hz = np.zeros(4, 'i4')

    #Offset_Hy[0] = NA(iMin - 1, jmin_hy, kmin_hy);
    #Offset_Hy[1] = NA(iMin,     jmin_hy, kmin_hy);
    #Offset_Hy[2] = NA(iMax - 1, jmin_hy, kmin_hy);
    #Offset_Hy[3] = NA(iMax,     jmin_hy, kmin_hy);
    offset_hy[0] = Ni * (iMin - 1) + Nj * jmin_hy + Nk * kmin_hy + N0
    offset_hy[1] = Ni * (iMin    ) + Nj * jmin_hy + Nk * kmin_hy + N0
    offset_hy[2] = Ni * (iMax - 1) + Nj * jmin_hy + Nk * kmin_hy + N0
    offset_hy[3] = Ni * (iMax    ) + Nj * jmin_hy + Nk * kmin_hy + N0

    #Offset_Hz[0] = NA(iMin - 1, jmin_hz, kmin_hz);
    #Offset_Hz[1] = NA(iMin,     jmin_hz, kmin_hz);
    #Offset_Hz[2] = NA(iMax - 1, jmin_hz, kmin_hz);
    #Offset_Hz[3] = NA(iMax,     jmin_hz, kmin_hz);
    offset_hz[0] = Ni * (iMin - 1) + Nj * jmin_hz + Nk * kmin_hz + N0
    offset_hz[1] = Ni * (iMin    ) + Nj * jmin_hz + Nk * kmin_hz + N0
    offset_hz[2] = Ni * (iMax - 1) + Nj * jmin_hz + Nk * kmin_hz + N0
    offset_hz[3] = Ni * (iMax    ) + Nj * jmin_hz + Nk * kmin_hz + N0

    length_hy = Nj * (jmax_hy - jmin_hy) + Nk * (kmax_hy - kmin_hy)
    length_hz = Nj * (jmax_hz - jmin_hz) + Nk * (kmax_hz - kmin_hz)

    #lbuf = length_hy + length_hz
    #print(Parm['comm_rank'], length_hy, length_hz, lbuf)
    #buf = np.zeros(lbuf, Parm['f_dtype'])

    return offset_hy, offset_hz, length_hy, length_hz

# 自分の領域はiを含むか(X方向)
def _mpi_proc(comm_size, comm_rank, Nx, iMin, iMax, i):

    return ((i >= iMin) and (i < iMax)) \
        or ((comm_rank == comm_size - 1) and (i == Nx))
"""
def mpi_proc(Parm, Nx, iMin, iMax):

    iProc = np.zeros(Nx + 1, 'i4')

    for i in range(Nx + 1):
        iProc[i] = ((i >= iMin) and (i < iMax)) \
            or ((Parm['comm_rank'] == Parm['comm_size'] - 1) and (i == Nx))
        #print(Parm['comm_size'], Parm['comm_rank'], i, iProc[i])

    return iProc
"""