/*
dftNear3d.c (OpenMP)

DFT of near field
*/

#include "ofd.h"

void dftNear3d(int itime)
{
	for (int ifreq = 0; ifreq < NFreq2; ifreq++) {
		const int64_t n0 = ifreq * NN;
		const int id = (itime * NFreq2) + ifreq;

		const real_t ef_r = (real_t)cEdft[id].r;
		const real_t ef_i = (real_t)cEdft[id].i;
		const real_t hf_r = (real_t)cHdft[id].r;
		const real_t hf_i = (real_t)cHdft[id].i;

		int i;

		// Ex
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin; i <  iMax; i++) {
		for (int j = jMin; j <= jMax; j++) {
		for (int k = kMin; k <= kMax; k++) {
			const int64_t n = NA(i, j, k);
			cEx_r[n0 + n] += Ex[n] * ef_r;
			cEx_i[n0 + n] += Ex[n] * ef_i;
		}
		}
		}

		// Ey
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin; i <= iMax; i++) {
		for (int j = jMin; j <  jMax; j++) {
		for (int k = kMin; k <= kMax; k++) {
			const int64_t n = NA(i, j, k);
			cEy_r[n0 + n] += Ey[n] * ef_r;
			cEy_i[n0 + n] += Ey[n] * ef_i;
		}
		}
		}

		// Ez
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin; i <= iMax; i++) {
		for (int j = jMin; j <= jMax; j++) {
		for (int k = kMin; k <  kMax; k++) {
			const int64_t n = NA(i, j, k);
			cEz_r[n0 + n] += Ez[n] * ef_r;
			cEz_i[n0 + n] += Ez[n] * ef_i;
		}
		}
		}

		// Hx
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin - 1; i <= iMax + 1; i++) {
		for (int j = jMin - 1; j <  jMax + 1; j++) {
		for (int k = kMin - 1; k <  kMax + 1; k++) {
			const int64_t n = NA(i, j, k);
			cHx_r[n0 + n] += Hx[n] * hf_r;
			cHx_i[n0 + n] += Hx[n] * hf_i;
		}
		}
		}

		// Hy
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin - 1; i <  iMax + 1; i++) {
		for (int j = jMin - 1; j <= jMax + 1; j++) {
		for (int k = kMin - 1; k <  kMax + 1; k++) {
			const int64_t n = NA(i, j, k);
			cHy_r[n0 + n] += Hy[n] * hf_r;
			cHy_i[n0 + n] += Hy[n] * hf_i;
		}
		}
		}

		// Hz
#ifdef _OPENMP
#pragma omp parallel for
#endif
		for (    i = iMin - 1; i <  iMax + 1; i++) {
		for (int j = jMin - 1; j <  jMax + 1; j++) {
		for (int k = kMin - 1; k <= kMax + 1; k++) {
			const int64_t n = NA(i, j, k);
			cHz_r[n0 + n] += Hz[n] * hf_r;
			cHz_i[n0 + n] += Hz[n] * hf_i;
		}
		}
		}

	}
}
