/*
setupSize.c

setup array size
*/

#include "ofd.h"
#include "ofd_prototype.h"

void setupSize(int comm_size, int comm_rank)
{
	// too many prosess
	if (comm_size > Nx) {
		if (comm_rank == 0) {
			fprintf(stderr, "*** too many processes = %d (limit = %d)\n", comm_size, Nx);
			fflush(stderr);
		}
	}

	// X region
	if (comm_size == 1) {
		iMin = 0;
		iMax = Nx;
	}
	else {
		const int nx = MAX(Nx / comm_size, 1);
		iMin = (comm_rank + 0) * nx;
		iMax = (comm_rank + 1) * nx;
		if (comm_rank == comm_size - 1) {
			iMax = Nx;
		}
	}

	jMin = 0;
	jMax = Ny;
	kMin = 0;
	kMax = Nz;
	//printf("%d %d %d %d %d %d %d %d\n", comm_size, comm_rank, iMin, iMax, jMin, jMax, kMin, kMax);

	// array index

	const int lx = (iABC == 0) ? 1 : (iABC == 1) ? cPML.l : 0;
	const int ly = (iABC == 0) ? 1 : (iABC == 1) ? cPML.l : 0;
	const int lz = (iABC == 0) ? 1 : (iABC == 1) ? cPML.l : 0;

	Nk = 1;
	Nj = ((kMax - kMin) + (2 * lz) + 1);
	Ni = ((jMax - jMin) + (2 * ly) + 1) * Nj;
	N0 = -((iMin - lx) * Ni + (jMin - ly) * Nj + (kMin - lz) * Nk);
	NN = NA(iMax + lx, jMax + ly, kMax + lz) + 1;
	//printf("%d %d %zd %zd %zd %zd %zd\n", comm_size, comm_rank, Ni, Nj, Nk, N0, NN);
	assert(NA(iMin - lx, jMin - ly, kMin - lz) == 0);
	assert(NA(iMax + lx, jMax + ly, kMax + lz) == NN - 1);
}


// ABC (array size)
void setupABCsize(void)
{
	numMurHx = numMurHy = numMurHz = 0;
	numPmlEx = numPmlEy = numPmlEz = numPmlHx = numPmlHy = numPmlHz = 0;

	if      (iABC == 0) {
		setupMurHx(0);
		setupMurHy(0);
		setupMurHz(0);
		//printf("%zd %zd %zd\n", numMurHx, numMurHy, numMurHz);
	}
	else if (iABC == 1) {
		setupPmlEx(0);
		setupPmlEy(0);
		setupPmlEz(0);
		setupPmlHx(0);
		setupPmlHy(0);
		setupPmlHz(0);
		//printf("%zd %zd %zd %zd %zd %zd\n", numPmlEx, numPmlEy, numPmlEz, numPmlHx, numPmlHy, numPmlHz);
	}
}
